#include  "lcd_6100.h"
#include  "f8x8.h"
#include  "typedefs.h"
#include  "Semaphor.h"


LCD6100_Color_t     LCD6100_Colors[1 << LCD6100_COLOR_TABLE_BITS]; // Farbtabelle, Colors[0] = Hintergrund = Fllfarbe, Color[1] = Vordergrund = Linienfarbe
UINT8               LCD6100_MemCtrl;
LCD6100_Flags_t     LCD6100_Flags;
LCD6100_Rect_t      LCD6100_Clip;            // Clipping Koordinaten, alles was ausserhalb ist wird nicht dargestellt

LCD6100_Point_t     LCD6100_Cursor;          // aktuelle Position fr Schriften
LCD6100_Rect_t      LCD6100_Window;          // aktuelles Fenster fr Textausgaben
INT8                LCD6100_FontFirstChar;   // erstes Zeichen im Font
INT8                LCD6100_FontLastChar;    // letztes Zeichen im Font
UINT8               LCD6100_FontWidth;       // fixed Zeichenbreite des Fonts
UINT8               LCD6100_FontHeight;      // Zeichenhhe
UINT8               LCD6100_FontBitsPixel;   // Farbtiefe des Fonts

Semaphor_t LCD6100_SemaPixel;
Semaphor_t LCD6100_SemaChar;
Semaphor_t LCD6100_SemaCommand;
Semaphor_t LCD6100_SemaStr;

void LCD6100_Reset(void);
UINT32 LCD6100_Read(UINT8 bitcount);
void LCD6100_Command(const INT8 cmd);
void LCD6100_Data(const INT8 data);
void LCD6100_Send(INT8 data, const INT8 cmd);
void LCD6100_DelayMs(INT16 delay);
void LCD6100_Ellipse(LCD6100_Coord_t x, LCD6100_Coord_t y, LCD6100_Coord_t a, LCD6100_Coord_t b);
void LCD6100_DoFillRect(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2, const LCD6100_Color_t color);
void LCD6100_Line(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2);
UINT8 LCD6100_FontRead(const INT16 index);
UINT8 LCD6100_DrawChar(INT8 c);
void LCD6100_Circle(LCD6100_Coord_t x, LCD6100_Coord_t y, LCD6100_Coord_t r);

void InitLCD6100(void)
{
  INT8 i;
  UINT32 d = 0;

  CreateSemaphor(&LCD6100_SemaChar,1);
  CreateSemaphor(&LCD6100_SemaStr,1);
  CreateSemaphor(&LCD6100_SemaPixel,1);
  CreateSemaphor(&LCD6100_SemaCommand,1);
  
  prc2 = 1;                   // Protected PortDirectionReg!
  LCD6100_DPort |= LCD6100_OUTPins;
  LCD6100_Port |= LCD6100_OUTPins;
  LCD6100_Reset();

  LCD6100_Command(LCD6100_SOFT_RESET);
  LCD6100_Command(LCD6100_SLEEP_OUT);
  LCD6100_Command(LCD6100_BOOSTER_ON);

  // wait for booster
  while (!(d & 0x80000000))
  {
    LCD6100_Command(LCD6100_DISPLAY_STATUS);
    d = LCD6100_Read(33);
  }

  LCD6100_Command(LCD6100_COLOR_INTERFACE);
  LCD6100_Data(LCD6100_COLOR_8_BIT);


  LCD6100_Command(LCD6100_COLOR_SET);
/*
  for (i = 0; i < 8; i++)
    LCD6100_Data(LCD6100_RED);                  // red
  for (i = 0; i < 8; i++)
    LCD6100_Data(LCD6100_GREEN);                // green
  for (i = 0; i < 4; i++)
    LCD6100_Data(LCD6100_BLUE);                 // blue
*/
  LCD6100_Data(0x00);                  // red
  LCD6100_Data(0x02);                  // red
  LCD6100_Data(0x04);                  // red
  LCD6100_Data(0x06);                  // red
  LCD6100_Data(0x09);                  // red
  LCD6100_Data(0x0B);                  // red
  LCD6100_Data(0x0D);                  // red
  LCD6100_Data(0x0F);                  // red

  LCD6100_Data(0x00);                  // green
  LCD6100_Data(0x02);                  // green
  LCD6100_Data(0x04);                  // green
  LCD6100_Data(0x06);                  // green
  LCD6100_Data(0x09);                  // green
  LCD6100_Data(0x0B);                  // green
  LCD6100_Data(0x0D);                  // green
  LCD6100_Data(0x0F);                  // green

  LCD6100_Data(0x00);                  // blue
  LCD6100_Data(0x04);                  // blue
  LCD6100_Data(0x0B);                  // blue
  LCD6100_Data(0x0F);                  // blue


  LCD6100_Command(LCD6100_MEM_CONTROL);
  LCD6100_MemCtrl = (1 << LCD6100_MEM_MX) | (0 << LCD6100_MEM_MY) | (1 << LCD6100_MEM_RGB) | (0 << LCD6100_MEM_VW) | (0 << LCD6100_MEM_LAO);
  LCD6100_Data(LCD6100_MemCtrl);

  LCD6100_SetRect(LCD6100_Clip, LCD6100_SCREEN_LEFT, LCD6100_SCREEN_TOP, LCD6100_SCREEN_RIGHT, LCD6100_SCREEN_BOTTOM);

  LCD6100_SetBkColor(LCD6100_SCREEN_COLOR);     // Hintergrund
  LCD6100_SetFgColor(LCD6100_BLACK);            // Vordergrund
  LCD6100_SetFrColor(LCD6100_GRAY);             // Rahmen
  LCD6100_SetShColor(LCD6100_SILVER);           // Schatten

  LCD6100_ClearScreen();

  LCD6100_Command(LCD6100_DISPLAY_ON);

  // wait for display on
  while (!(d & 0x00000400))
  {
    LCD6100_Command(LCD6100_DISPLAY_STATUS);
    d = LCD6100_Read(33);
  }

  LCD6100_ClearScreen();
  

  // Font Init:
  LCD6100_FontWidth = LCD6100_FontRead(2);
  LCD6100_FontHeight = LCD6100_FontRead(3);
  LCD6100_FontBitsPixel = LCD6100_FontRead(4);
  LCD6100_FontFirstChar = LCD6100_FontRead(5);
  LCD6100_FontLastChar = LCD6100_FontRead(6);
  LCD6100_SetRect(LCD6100_Window, LCD6100_SCREEN_LEFT+2, LCD6100_SCREEN_TOP+2, LCD6100_SCREEN_RIGHT-2, LCD6100_SCREEN_BOTTOM-2);  
  LCD6100_Flags.AutoLineFeed = 1;
//  LCD6100_Flags.FixedFont = 1;

  LCD6100_Cursor.Y = 2;
  LCD6100_Cursor.X = 2;
  

/*

  // Test?

  LCD6100_Command(LCD6100_INVERSION_OFF);
  LCD6100_Command(LCD6100_INVERSION_ON);

  for (i = 1; i < 130; i++)
  {
    LCD6100_SetPixel(4,i,i);
  }

  for (i = 130; i < 255; i++)
  {
    LCD6100_SetPixel(5,i-130,i);
  }
  LCD6100_Print("ES Geht! Wegen einem Typecast! lol\n");

  LCD6100_SetPixel(50,50,LCD6100_GREEN);
  LCD6100_SetPixel(51,50,LCD6100_GREEN);
  LCD6100_SetPixel(52,50,LCD6100_GREEN);
  LCD6100_SetPixel(53,50,LCD6100_GREEN);
  LCD6100_SetPixel(54,50,LCD6100_GREEN);
  LCD6100_SetPixel(55,50,LCD6100_GREEN);
  LCD6100_SetPixel(56,50,LCD6100_GREEN);
  LCD6100_SetPixel(57,50,LCD6100_RED);
  LCD6100_DoFillRect(30, 30, 33, 33,   LCD6100_GREEN);
  LCD6100_DoFillRect(20, 20, 40, 40,   LCD6100_GOLD);
  LCD6100_DoFillRect(80, 80, 100, 100, LCD6100_BLUE);
  LCD6100_DoFillRect(90, 90, 95, 95,   LCD6100_MAGENTA);
  LCD6100_DoFillRect(52, 50, 52, 76,   LCD6100_GOLD);
  LCD6100_Line(5, 5, 125, 125);
  
//  LCD6100_ClearScreen();
*/
  LCD6100_Circle(50,50,20);
  
}

UINT32 LCD6100_Read(UINT8 bitcount)
{
  UINT32 data = 0;
  LCD6100_DelayMs(1);
  prc2 = 1;                                   // Protected PortDirectionReg!
  LCD6100_DPort &= ~LCD6100_DATABit;
  for (; bitcount > 0; bitcount--)
  {
    LCD6100_Port &= ~LCD6100_CLKBit;
    data += data;
    LCD6100_Port |= LCD6100_CLKBit;
    if (LCD6100_Port & LCD6100_DATABit) { data++; }
  }
  LCD6100_Port |= LCD6100_CSBit;              // wir mssen CS zurcksetzen !!
  prc2 = 1;                                   // Protected PortDirectionReg!
  LCD6100_DPort |= LCD6100_DATABit;
  LCD6100_Port &= ~LCD6100_CSBit;              
  return data;
}


void LCD6100_Command(const INT8 cmd) 
{
  P(&LCD6100_SemaCommand);
  LCD6100_Port &= ~LCD6100_CSBit;              // CS 0 => LCD Selected
  LCD6100_Send(cmd, 1);
  V(&LCD6100_SemaCommand);
}

void LCD6100_Data(const INT8 data)
{
  P(&LCD6100_SemaPixel);
  LCD6100_Send(data, 0);
  V(&LCD6100_SemaPixel);
}

/* Transfer von data zum GLCD.
   Der PCF8833 benutzt ein 9Bit Telegram bei dem das MSB angibt ob man Daten oder ein Kommando sendet. */
void LCD6100_Send(char data, const INT8 cmd) 
{
  UINT8 i;
  LCD6100_Port &= ~LCD6100_CLKBit;
  if (cmd) 
  {
    LCD6100_Port &= ~LCD6100_DATABit;
  } else 
  {
    LCD6100_Port |= LCD6100_DATABit;
  }
  LCD6100_Port |= LCD6100_CLKBit;
  
  // Send 8 data-bits
  for (i = 8; i != 0; i--)
  {
    LCD6100_Port &= ~LCD6100_CLKBit;
    if (data & 0x80) 
    {
      LCD6100_Port |= LCD6100_DATABit;
    } else 
    {
      LCD6100_Port &= ~LCD6100_DATABit;
    }
    LCD6100_Port |= LCD6100_CLKBit;      
    data = data << 1;
  }
}

void LCD6100_DelayMs(int delay)
{
  delay = delay << 8;
  while (delay > 0) delay--;
}

void LCD6100_Reset(void)
{
  LCD6100_DelayMs(30);
  LCD6100_Port &= ~LCD6100_RESETBit;
  LCD6100_Port |=  LCD6100_RESETBit;
}


/* setze Display Ausrichtung, 0,90,180,270
   ACHTUNG! bei 90 und 270 werden alle X,Y Koordinaten vertauscht aber nicht die absoluten SCREEN_xxxx Koordinaten.
   Auf einem 132x132 Nokia Display hat das keinerlei Konsequenzen, aber auf einem zB. 176x240 Display wrde das nicht
   funktionieren da dann das Clipping nicht richtig wre. Auf so einem Display mssen die Clipping Rects Clip und Window
   ebenfalls angepasst werden. 
   Das Hardware Scrolling des PCF8833 kann dieses virtuelle drehen der Ausrichtung ebenfalls nicht !!*/
void LCD6100_SetOrientation(const UINT8 orientation)
{
  LCD6100_Command(LCD6100_MEM_CONTROL);
  LCD6100_MemCtrl &= ~((1 << LCD6100_MEM_90) | (1 << LCD6100_MEM_VW) | (1 << LCD6100_MEM_MX) | (1 << LCD6100_MEM_MY));
  if (orientation & 1) {
    LCD6100_MemCtrl |= (1 << LCD6100_MEM_90) | (1 << LCD6100_MEM_VW) | (1 << LCD6100_MEM_MX);
  }
  if (orientation & 2) {
    LCD6100_MemCtrl ^= (1 << LCD6100_MEM_MY);
  } else {
    LCD6100_MemCtrl ^= (1 << LCD6100_MEM_MX);
  }
  LCD6100_Data(LCD6100_MemCtrl);
}

void LCD6100_SetAddr(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2)
{
 if (LCD6100_MemCtrl & (1 << LCD6100_MEM_90))
 {
    LCD6100_SwapCoord(x1, y1);
    LCD6100_SwapCoord(x2, y2);
  }
  LCD6100_Command(LCD6100_SET_X_ADDR);
  LCD6100_Data(x1);
  LCD6100_Data(x2);
  LCD6100_Command(LCD6100_SET_Y_ADDR);
  LCD6100_Data(y1);
  LCD6100_Data(y2);
  LCD6100_Command(LCD6100_MEM_WRITE);
}

void LCD6100_SetPixel(LCD6100_Coord_t x, LCD6100_Coord_t y, const LCD6100_Color_t color)
{
  if ((x >= LCD6100_Clip.X1) & (x <= LCD6100_Clip.X2) & (y >= LCD6100_Clip.Y1) & (y <= LCD6100_Clip.Y2))
  {
    LCD6100_SetAddr(x, y, LCD6100_SCREEN_WIDTH -1, LCD6100_SCREEN_HEIGHT -1);
    LCD6100_PutPixel(color);
  }
}

void LCD6100_DoFillRect(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2, const LCD6100_Color_t color)
{
  LCD6100_Coord_t x, y;
  if ((x1 > LCD6100_Clip.X2) | (x2 < LCD6100_Clip.X1) | (y1 > LCD6100_Clip.Y2) | (y2 < LCD6100_Clip.Y1)) { return; }
  if (x1 < LCD6100_Clip.X1) { x1 = LCD6100_Clip.X1; }
  if (y1 < LCD6100_Clip.Y1) { y1 = LCD6100_Clip.Y1; }
  if (x2 > LCD6100_Clip.X2) { x2 = LCD6100_Clip.X2; }
  if (y2 > LCD6100_Clip.Y2) { y2 = LCD6100_Clip.Y2; }
  if ((x1 <= x2) & (y1 <= y2))
  {
    LCD6100_SetAddr(x1, y1, x2, y2);
    for (y = y1; y <= y2; y++)
    {
      for (x = x1; x <= x2; x++)
      {
        LCD6100_PutPixel(color);
      }
    }
  }
}

void LCD6100_FillRect(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2, const LCD6100_Color_t color)
{
  LCD6100_CheckCoord(x1, x2);
  LCD6100_CheckCoord(y1, y2);
  LCD6100_DoFillRect(x1, y1, x2, y2, color);
}

void LCD6100_Rectangle(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2)
{
  LCD6100_CheckCoord(x1, x2);
  LCD6100_CheckCoord(y1, y2);
  if (LCD6100_FgColor != LCD6100_NONE)
  {
    LCD6100_DoFillRect(x1, y1, x2, y1, LCD6100_FgColor);
    LCD6100_DoFillRect(x2, y1, x2, y2, LCD6100_FgColor);
    LCD6100_DoFillRect(x1, y2, x2, y2, LCD6100_FgColor);
    LCD6100_DoFillRect(x1, y1, x1, y2, LCD6100_FgColor);
  }
  if (LCD6100_BkColor != LCD6100_NONE)
  {
    LCD6100_DoFillRect(++x1, ++y1, --x2, --y2, LCD6100_BkColor);
  }
}

void LCD6100_Frame(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2)
{
  LCD6100_CheckCoord(x1, x2);
  LCD6100_CheckCoord(y1, y2);

  if (LCD6100_ShColor != LCD6100_NONE)
  {
    LCD6100_DoFillRect(x1 + 4, y2, x2 - 2, y2, LCD6100_ShColor);
    LCD6100_DoFillRect(x2, y1 +4, x2, y2 -2, LCD6100_ShColor);
    x2--;
    y2--;
    LCD6100_DoFillRect(x2 -2, y2 -2, x2, y2, LCD6100_ShColor);
  }
  if (LCD6100_FrColor != LCD6100_NONE)
  {
    LCD6100_DoFillRect(x1 +2, y1, x2 -2, y1, LCD6100_FrColor);
    LCD6100_DoFillRect(x1 +2, y2, x2 -2, y2, LCD6100_FrColor);
    LCD6100_DoFillRect(x2, y1 +2, x2, y2 -2, LCD6100_FrColor);
    LCD6100_DoFillRect(x1, y1 +2, x1, y2 -2, LCD6100_FrColor);
    x1++;
    y1++;
    x2--;
    y2--;
    LCD6100_SetPixel(x1, y1, LCD6100_FrColor);
    LCD6100_SetPixel(x2, y1, LCD6100_FrColor);
    LCD6100_SetPixel(x1, y2, LCD6100_FrColor);
    LCD6100_SetPixel(x2, y2, LCD6100_FrColor);
  }
  if (LCD6100_BkColor != LCD6100_NONE)
  {
    LCD6100_DoFillRect(x1, y1 +1, x1, y2 -1, LCD6100_BkColor);
    LCD6100_DoFillRect(x2, y1 +1, x2, y2 -1, LCD6100_BkColor);
    LCD6100_DoFillRect(x1 +1, y1, x2 -1, y2, LCD6100_BkColor);
  }
}

void LCD6100_Line(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2)
{
  UINT8 swap;
  INT16 e;
  LCD6100_Coord_t x3;
  LCD6100_Coord_t y3;
  LCD6100_Coord_t i;
  
  if (LCD6100_FgColor != LCD6100_NONE)
  {
    if ((x1 == x2) | (y1 == y2)) {
      LCD6100_FillRect(x1, y1, x2, y2, LCD6100_FgColor);
    } else
    {

      LCD6100_Coord_t dx, dy, sx, sy;

      if (x1 < x2) {dx = x2 - x1; sx = +1;}
        else {dx = x1 - x2; sx = -1;}
      if (y1 < y2) {dy = y2 - y1; sy = +1;}
        else {dy = y1 - y2; sy = -1;}

      swap = (dx > dy) ? 0 : 1;
      if (swap) {
        LCD6100_SwapCoord(x1, y1);
        LCD6100_SwapCoord(dx, dy);
        LCD6100_SwapCoord(sx, sy);
      }

      e = dy - dx + dy;
      x3 = x1;
      y3 = y1;
      x2 = x1;
      y2 = y1;
/* Diese Loop ist suboptimal. Man knnte mehr rausholen wenn man die Hardware besser ausnutzen wrde, aber
 dann wre diese Routine eben nicht mehr Hardwareunabhngig so wie jetzt. */
      for (i = 0; i <= dx; i++)
      {
        if ((x1 != x2) & (y1 != y2))
        {
          if (swap) {
            LCD6100_SwapCoord(x2, y2);
            LCD6100_SwapCoord(x3, y3);
          }
          LCD6100_FillRect(x2, y2, x3, y3, LCD6100_FgColor);
          x2 = x1;
          y2 = y1;
        };
        x3 = x1;
        y3 = y1;
        while (e >= 0) {
          y1 += sy;
          e  -= dx;
          e  -= dx;
        }
        x1 += sx;
        e  += dy;
        e  += dy;
      }
      if (swap) {
        LCD6100_SwapCoord(x2, y2);
        LCD6100_SwapCoord(x3, y3);
      }
      LCD6100_FillRect(x2, y2, x3, y3, LCD6100_FgColor);
    }
  }
}

/* Hilfsfunktion fr Ellipsen */
void LCD6100_DoPixelLine(int x1, INT16 x2, const INT16 y, const UINT8 fill)
{
  if ((y >= LCD6100_Clip.Y1 ) & (y <= LCD6100_Clip.Y2)) {
    if (x1 < LCD6100_Clip.X1) {x1 = LCD6100_Clip.X1;} else {
      if (x1 > LCD6100_Clip.X2) {return;} else {
        if (LCD6100_FgColor != LCD6100_NONE) {
          LCD6100_SetPixel(x1, y, LCD6100_FgColor);
        }
        x1++;
      }
    }
    if (x2 < LCD6100_Clip.X1) {return;} else {
      if (x2 > LCD6100_Clip.X2) {x2 = LCD6100_Clip.X2;} else {
        if (LCD6100_FgColor != LCD6100_NONE) {
          LCD6100_SetPixel(x2, y, LCD6100_FgColor);
        }
        x2--;
      }
    }
    if ((fill) & (LCD6100_BkColor != LCD6100_NONE) & (x1 <= x2)) {
      LCD6100_DoFillRect(x1, y, x2, y, LCD6100_BkColor);
    }
  }
}

void LCD6100_Circle(LCD6100_Coord_t x, LCD6100_Coord_t y, LCD6100_Coord_t r)
{
  LCD6100_Ellipse(x, y, r, r);
}

void LCD6100_Ellipse(LCD6100_Coord_t x, LCD6100_Coord_t y, LCD6100_Coord_t a, LCD6100_Coord_t b)
{
  LCD6100_Coord2_t aa;
  LCD6100_Coord2_t bb;
  INT32 er, cr, ir;
  INT16 ys, ye,xs, xe;
  UINT8 fill;

  if ((a == 0) | (b == 0)) {return;}
  if ((a > 180) | (b > 180)) {
    if (LCD6100_BkColor != LCD6100_NONE)
    {
      LCD6100_FillRect(0, 0, LCD6100_SCREEN_WIDTH -1, LCD6100_SCREEN_HEIGHT -1, LCD6100_BkColor);
     }
     return;
  }
  
  aa = a * a;
  bb = b * b;

  cr = bb >> 1;
  cr = cr * (a + a -1);
  ir = aa >> 1;
  ir = -ir;
  er = 0;

  xs = x;
  xs = xs - a;
  xe = x;
  xe = xe + a;
  ys = y;
  ye = y;
  while (cr >= ir)
  {
    LCD6100_DoPixelLine(xs, xe, ys, 1);
    if (ys != ye) 
    {
      LCD6100_DoPixelLine(xs, xe, ye, 1);
    }
    ys--;
    ye++;
    ir += aa;
    er += ir;
    if (2 * er > cr) 
    {
      er -= cr;
      cr -= bb;
      xs++;
      xe--;
    }
  }

  cr = aa >> 1;
  cr = cr * (b + b -1);
  ir = bb >> 1;
  ir = -ir;
  er = 0;

  xs = x;
  xe = x;
  ys = y;
  ys = ys - b;
  ye = y;
  ye = ye + b;
  fill = 1;
  while (ir <= cr)
  {
    LCD6100_DoPixelLine(xs, xe, ys, fill);
    if (ys != ye) 
    {
      LCD6100_DoPixelLine(xs, xe, ye, fill);
    }
    fill = 0;
    ir += bb;
    er += ir;
    if (2 * er > cr)
    {
      er -= cr;
      cr -= aa;
      ys++;
      ye--;
      fill = 1;
    }
    xs--;
    xe++;
  }
}

UINT8 LCD6100_FontRead(const INT16 index)
{
  return f8x8[index];
//  return(pgm_read_byte(index));
}

void LCD6100_NewLine(void)
{
  LCD6100_Cursor.X = LCD6100_Window.X1;
  LCD6100_Cursor.Y += LCD6100_FontHeight;
  if (LCD6100_Cursor.Y + LCD6100_FontHeight > LCD6100_Window.Y2)
  {
    LCD6100_Cursor.Y = LCD6100_Window.Y1;
  }
}

UINT8 LCD6100_DrawChar(char c)
{
  UINT8 charwidth;
  UINT8 width;
  UINT8 i;

  LCD6100_FontData_t data;                                // *Font Daten
  UINT8 bitspixel;                                        // Bits pro Pixel
  UINT8 bitsmask;                                         // Bitmaske fr Farbe
  UINT16  abits;                                          // enthlt die aktuelle Bits des Zeichens
  UINT8 bitscount;                                        // aktuelle Anzahl an Bits in Bits
  UINT8 table[4];                                         // RLE Lngen Tabelle
  
  UINT8 padding;

  UINT8 pixelcount;                                       // Anzahl der zu setzenden Pixel
  LCD6100_Color_t pixelcolor;                             // zu setzende Pixelfarbe
  UINT8 clipinvalid;                                      // Gltigkeit der Adresszeiger im PCF8833
  LCD6100_Coord_t clipY;                                  // letzte gesetzte Y-Adresse im PCF8833
  LCD6100_Coord_t stopY;                                  // unterste Pixelzeile des Zeichens
  LCD6100_Coord_t stopX;
  
  LCD6100_Coord_t x;
  LCD6100_Coord_t y;
  
  UINT32 index;
  
  
  P(&LCD6100_SemaChar);

  if (c == '\n')
  {
    LCD6100_NewLine();
    goto __return0;
  }
  
  // ist das Zeichen im Font definiert und auch ein Font selektiert ??
  if ((c < LCD6100_FontFirstChar) | (c > LCD6100_FontLastChar))
  {
    goto __return0;
  }

  // ist das Zeichen im Font deklariert aber nicht definiert ??
  charwidth = LCD6100_FontRead(LCD6100_FONT_HEADER_SIZE + c - LCD6100_FontFirstChar);
  if (charwidth == 0) { goto __return0; }

  // alle Zeichen kleiner #128 haben rechts eine implizite Leerspalte
  width = charwidth;
  if (c < 128) { width++; }

  // wenn wir den Zeichensatz mit fester Breite zeichnen sollen dann setze die Width entsprechend
  if ((LCD6100_Flags.FixedFont == 1) & (width <= LCD6100_FontWidth))
  {
    width = LCD6100_FontWidth;
  }

  // passt das Zeichen noch in das Zeichen-Fenster ?? falls nein eventuell ein Linefeed einfgen
  if ((LCD6100_Flags.AutoLineFeed == 1) & (LCD6100_Cursor.X + width >= LCD6100_Window.X2))
  {
    LCD6100_NewLine();
  }

  // ist das Zeichen berhaupt noch ansatzweise sichtbar ?
  if ((LCD6100_Cursor.X > LCD6100_Clip.X2) | (LCD6100_Cursor.Y > LCD6100_Clip.Y2))
  {
    goto __return0;
  }

  x = LCD6100_Cursor.X;
  y = LCD6100_Cursor.Y;
  LCD6100_Cursor.X += width;

  // ist das Zeichen berhaupt noch ansatzweise sichtbar ?
  if ((x > LCD6100_Clip.X2) | (y + LCD6100_FontHeight < LCD6100_Clip.Y1))
  {
    goto __return0;
  }

  i = 0;

  // falls FixedFont mssen die Zeichen innerhalb LCD6100_FontWidth zentriert werden, aber nur
  // wenn das Zeichen schmaler als LCD6100_FontWidth ist. Wir fllen dann auch gleich den rechten
  // und linken Rand des Zeichens, ebenfalls nur wenn die Hintergrundfarbe nicht transparent ist.
  if ((LCD6100_Flags.FixedFont == 1) & (width > charwidth))
  {
    i = (width - charwidth) / 2;
    if (LCD6100_BkColor != LCD6100_NONE)
    {
      LCD6100_DoFillRect(x, y, x + i -1, y + LCD6100_FontHeight -1, LCD6100_BkColor);
    }
    x += i;
  }
  i = width - charwidth - i;
  if ((i > 0) & (LCD6100_BkColor != LCD6100_NONE))
  {
    LCD6100_DoFillRect(x + charwidth, y, x + charwidth + i -1, y + LCD6100_FontHeight -1, LCD6100_BkColor);
  }

  // mssen die eigentlichen Zeichendaten gezeichnet werden ??
  if ((x > LCD6100_Clip.X2) | (x + charwidth < LCD6100_Clip.X1))
  {
    goto __return;
  }

  bitspixel = LCD6100_FontBitsPixel & 0x7F;             // Bits pro Pixel
  bitsmask = 0xFF >> (8 - bitspixel);                // Bitmaske fr Farbe

  // ALLE Lesevorgnge in den Font-Speicher sind so aufgebaut das sie ganze Blcke von
  // Bytes sequentiell erledigen. Dies ist wichtig und sehr gut optimiert auf zB. I2C EEPROMS.

  // berechne nun den Bitindex zu den Bitdaten im Font
  data =  LCD6100_FONT_HEADER_SIZE;
  index = 0;
  for (i = LCD6100_FontFirstChar; i < c; i++) {
    index += LCD6100_FontRead(data++);
  }
  
  // linkes Clipping bercksichtigen, nur bei unkomprimierten Fonts mglich
  if (x < LCD6100_Clip.X1) {
    LCD6100_Coord_t t = LCD6100_Clip.X1 - x;
    index += t;
    charwidth -= t;
    x = LCD6100_Clip.X1;
  }
  index *= LCD6100_FontHeight * bitspixel;

  // positioniere data auf's erste Byte mit Fontdaten zum Zeichen und hole die ersten Bits
  bitscount = index % 8;
  index /= 8;
  data = LCD6100_FONT_HEADER_SIZE + LCD6100_FontLastChar - LCD6100_FontFirstChar +1 + index;
  abits = LCD6100_FontRead(data++) >> bitscount;
  bitscount = 8 - bitscount;

  // wir senden die Bitdaten der einzelnen Pixel jeweils Spaltenweise in Y Richtung
  LCD6100_Command(LCD6100_MEM_CONTROL);
  LCD6100_MemCtrl ^= (1 << LCD6100_MEM_VW);
  LCD6100_Data(LCD6100_MemCtrl);

  pixelcount =  0;                                      // Anzahl der zu setzenden Pixel
  pixelcolor =  0;                                      // zu setzende Pixelfarbe
  clipinvalid = 1;                                      // Gltigkeit der Adresszeiger im PCF8833
  clipY = 0;                                            // letzte gesetzte Y-Adresse im PCF8833
  stopY = LCD6100_Cursor.Y + LCD6100_FontHeight -1;     // unterste Pixelzeile des Zeichens
  stopX = x + charwidth -1;

  // rechtes Clipping bercksichtigen
  if (stopX > LCD6100_Clip.X2)
  {
    stopX = LCD6100_Clip.X2;
  }

  for (; x <= stopX; x++)
  {
    if (clipY |= LCD6100_Cursor.Y)
    {
      clipinvalid = 1;
    }
    for (y = LCD6100_Cursor.Y; y <= stopY; y++)
    {
      if (pixelcount == 0)
      {
        if (bitscount <= 8)
        {
          abits |= ((UINT16) LCD6100_FontRead(data++)) << bitscount;
          bitscount += 8;
        }
        if (LCD6100_FontBitsPixel & 0x80)
        {
          pixelcount = table[abits & 3];
          abits >>= 2;
          bitscount -= 2;
        } else
        {
          pixelcount++;
        }
        if (bitspixel <= LCD6100_COLOR_TABLE_BITS)
        {
          pixelcolor = LCD6100_Colors[abits & bitsmask];
        } else
        {
          pixelcolor = abits & bitsmask;
        }
        abits >>= bitspixel;
        bitscount -= bitspixel;
      }
      pixelcount--;
      if ((pixelcolor != LCD6100_NONE) && (y >= LCD6100_Clip.Y1) && (y <= LCD6100_Clip.Y2) && (x >= LCD6100_Clip.X1))
      {
        if (clipinvalid)
        {
          LCD6100_SetAddr(x, y, LCD6100_Cursor.X, stopY);
          clipY = y;
          clipinvalid = 0;
        }
        LCD6100_PutPixel(pixelcolor);
      } else
      {
        clipinvalid = 1;
      }
    }
  }

  LCD6100_Command(LCD6100_MEM_CONTROL);
  LCD6100_MemCtrl ^= (1 << LCD6100_MEM_VW);
  LCD6100_Data(LCD6100_MemCtrl);

__return0:  
  V(&LCD6100_SemaChar);
  return(0);
__return:  
  V(&LCD6100_SemaChar);
  return(width);
}

UINT8 LCD6100_CharWidth(INT8 c)
{
  if ((c < LCD6100_FontFirstChar) | (c > LCD6100_FontLastChar))
  {
    return(0); // character not defined
  } else
  {
    UINT8 i = LCD6100_FontRead(LCD6100_FONT_HEADER_SIZE + c - LCD6100_FontFirstChar);
    if (i > 0)
    {
      if (c < 128)
      {
        i++;
      }
      if ((LCD6100_Flags.FixedFont) & (i < LCD6100_FontWidth))
      {
        i = LCD6100_FontWidth;
      }
    }
    return(i);
  }
}

UINT16 LCD6100_Print(UINT8 far *addr)
{
  UINT8 c;
  UINT16 i = 0;
  P(&LCD6100_SemaStr);
  while ((c = *addr++))
  {
    i += LCD6100_DrawChar(c);
  }
  V(&LCD6100_SemaStr);
  return(i);
}

UINT16 LCD6100_Print_P(UINT16 addr)
{
  UINT8 c;
  UINT16 i = 0;
  while ((c = LCD6100_FontRead(addr++)))
  {
    i += LCD6100_DrawChar(c);
  }
  return(i);
}

UINT16 LCD6100_CharsWidth(const INT8 *addr)
{
  UINT8 c;
  UINT16 i = 0;

  while ((c = *addr++))
  {
    i += LCD6100_CharWidth(c);
  }
  return(i);
}

UINT16 LCD6100_CharsWidth_P(UINT16 addr)
{
  UINT8 c;
  UINT16 i = 0;

  while ((c = LCD6100_FontRead(addr++)))
  {
    i += LCD6100_CharWidth(c);
  }
  return(i);
}
