#ifndef __LCD_6100H_
#define __LCD_6100H_

#include  <stdio.h>
#include  "sfr62p.h"
#include  "CPU.h"
#include  "UART0.h"
#include  "typedefs.h"


#define LCD6100_WidthPixel   132
#define LCD6100_HeightPixel  132

#define LCD6100_RESETBit (1 << 4)        // Reset-Pin
#define LCD6100_DATABit  (1 << 5)        // Data
#define LCD6100_CLKBit   (1 << 6)        // Clock bit
#define LCD6100_CSBit    (1 << 7)        // Chip select bit

#define LCD6100_OUTPins  (LCD6100_CSBit|LCD6100_CLKBit|LCD6100_DATABit|LCD6100_RESETBit)	// Output Config Pins

#define LCD6100_Port     p9
#define LCD6100_DPort    pd9

#define LCD6100_BufferSize  ((LCD6100_WidthPixel * LCD6100_HeightPixel) / 8)

#define LCD6100_SCREEN_COLOR            LCD6100_WHITE  // sollte gendert werden fr andere Hintergrundfarbe
#define LCD6100_COLOR_TABLE_BITS        2      // 2^2 = 4 Farben in Farbtabelle, wichtig fr Fonts
#define LCD6100_FONT_HEADER_SIZE        7      // Header gre eines Fonts

typedef UINT8   LCD6100_Coord_t;          // Koordinaten
typedef UINT16  LCD6100_Coord2_t;        // intern zur Koordinatenberechnung
typedef UINT8   LCD6100_Color_t;          // Farbe, fr 256 Farbe uint8_t
typedef UINT16  LCD6100_FontData_t;      // zeiger auf Fontdaten

typedef struct {
    LCD6100_Coord_t X;
    LCD6100_Coord_t Y;
} LCD6100_Point_t;

typedef struct {
    LCD6100_Coord_t X1;
    LCD6100_Coord_t Y1;
    LCD6100_Coord_t X2;
    LCD6100_Coord_t Y2;
} LCD6100_Rect_t;

typedef struct {
    UINT8
            AutoLineFeed:1,        // Fontroutine nutzt einen impliziten Zeilenumbruch innerhalb von glcd_Window
            FixedFont:1;           // Fontroutine soll Font mit fester Zeichenbreite darstellen, ansonsten proportional
} LCD6100_Flags_t;

#define LCD6100_SCREEN_WIDTH            132   // Nokia6100, wobei aber nur 1,1-130,130 sichtbar sind. Denoch bentigen wir vollen
#define LCD6100_SCREEN_HEIGHT           132   // Zugriff auf den RAM wenn wir zB. das Scrolling nutzen wollen
#define LCD6100_SCREEN_LEFT             1
#define LCD6100_SCREEN_TOP              1
#define LCD6100_SCREEN_RIGHT            (LCD6100_SCREEN_WIDTH -2)
#define LCD6100_SCREEN_BOTTOM           (LCD6100_SCREEN_HEIGHT -2)

#define LCD6100_RGB(r,g,b) ((r & 0xE0) | ((g & 0xE0) >> 3) | (b >> 6))

#define LCD6100_NONE                    LCD6100_RGB(0x00, 0x20, 0x00)  // == tansparent, ein grn Ton weniger !!
#define LCD6100_BLACK                   LCD6100_RGB(0x00, 0x00, 0x00)  // fr 8 Bit Farben R,G,B  3,3,2 Bits
#define LCD6100_WHITE                   LCD6100_RGB(0xFF, 0xFF, 0xFF)
#define LCD6100_RED                     LCD6100_RGB(0xFF, 0x00, 0x00)
#define LCD6100_GREEN                   LCD6100_RGB(0x00, 0xFF, 0x00)
#define LCD6100_BLUE                    LCD6100_RGB(0x00, 0x00, 0xFF)
#define LCD6100_YELLOW                  LCD6100_RGB(0xFF, 0xFF, 0x00)
#define LCD6100_MAGENTA                 LCD6100_RGB(0xFF, 0x00, 0xFF)
#define LCD6100_CYAN                    LCD6100_RGB(0x00, 0xFF, 0xFF)
#define LCD6100_GRAY                    LCD6100_RGB(0x80, 0x80, 0x40)
#define LCD6100_SILVER                  LCD6100_RGB(0xA0, 0xA0, 0x80)
#define LCD6100_GOLD                    LCD6100_RGB(0xA0, 0xA0, 0x40)

#define LCD6100_BkColor LCD6100_Colors[0]
#define LCD6100_FgColor LCD6100_Colors[1]
#define LCD6100_FrColor LCD6100_Colors[2]
#define LCD6100_ShColor LCD6100_Colors[3]
#define LCD6100_SetColor(index, color) {LCD6100_Colors[index] = color;}
#define LCD6100_SetBkColor(color) {LCD6100_SetColor(0, color);}
#define LCD6100_SetFgColor(color) {LCD6100_SetColor(1, color);}
#define LCD6100_SetFrColor(color) {LCD6100_SetColor(2, color);}
#define LCD6100_SetShColor(color) {LCD6100_SetColor(3, color);}
#define LCD6100_SetColors(fgColor, bkColor) {LCD6100_SetFgColor(fgColor); LCD6100_SetBkColor(bkColor);}
#define LCD6100_PutPixel(color) {LCD6100_Data(color);}
#define LCD6100_ClearScreen() {LCD6100_FillRect(0, 0, LCD6100_SCREEN_WIDTH -1, LCD6100_SCREEN_HEIGHT -1, LCD6100_SCREEN_COLOR);}
#define LCD6100_SetRect(r,x1,y1,x2,y2) {r.X1 = x1; r.Y1 = y1; r.X2 = x2; r.Y2 = y2;}
#define LCD6100_SwapCoord(x, y) {LCD6100_Coord_t t = x; x = y; y = t;}
#define LCD6100_CheckCoord(x, y) {if (x > y) {LCD6100_SwapCoord(x, y)}}

// Commands

#define LCD6100_NO_OPERATION            0x00
#define LCD6100_SOFT_RESET              0x01
#define LCD6100_BOOSTER_OFF             0x02
#define LCD6100_BOOSTER_ON              0x03
#define LCD6100_DISPLAY_INFO            0x04   // read 25bits where MSB are dummy trash
#define LCD6100_DISPLAY_STATUS          0x09   // read 33bits where MSB are dummy trash
#define LCD6100_SLEEP_IN                0x10
#define LCD6100_SLEEP_OUT               0x11
#define LCD6100_PARTIAL_MODE_ON         0x12
#define LCD6100_NORMAL_MODE_ON          0x13
#define LCD6100_INVERSION_OFF           0x20
#define LCD6100_INVERSION_ON            0x21
#define LCD6100_PIXELS_OFF              0x22
#define LCD6100_PIXELS_ON               0x23
#define LCD6100_SET_CONTRAST            0x25
  // CONTRAST_VALE              0xXX
#define LCD6100_DISPLAY_OFF             0x28
#define LCD6100_DISPLAY_ON              0x29
#define LCD6100_SET_X_ADDR              0x2A  // ATENTION! when X,Y address are changed BOTH must be changed
  // START_X                    0xXX
  // STOP_X                     0xXX
#define LCD6100_SET_Y_ADDR              0x2B
  // START_Y                    0xXX
  // STOP_Y                     0xXX
#define LCD6100_MEM_WRITE               0x2C
  // x times pixel data         0xXX
#define LCD6100_COLOR_SET               0x2D
  // 8 times red palette        0xXX
  // 8 times green palette      0xXX
  // 4 times blue palette       0xXX
#define LCD6100_PARTIAL_AREA            0x30
  // START_Y                    0xXX
  // STOP_Y                     0xXX
#define LCD6100_SCROLL_DEFINITON        0x33  // always vertical
  // TOP_FIXED_LINES            0xXX
  // SCROLL_LINES               0xXX
  // BOTTOM_FIXED_LINES         0xXX
#define LCD6100_SET_SCROLL              0x37  // SEP scroll entry point
  // ROW_Y                      0xxx
#define LCD6100_MEM_CONTROL             0x36
  // MEM_FLAGS                  0xXX
  #define LCD6100_MEM_90                0     // virtual flag not for PCF8833
  #define LCD6100_MEM_RGB               3     // color order
  #define LCD6100_MEM_LAO               4     // line access order
  #define LCD6100_MEM_VW                5     // vertical write
  #define LCD6100_MEM_MX                6     // mirror X
  #define LCD6100_MEM_MY                7     // mirror Y
#define LCD6100_IDLE_OFF                0x38
#define LCD6100_IDLE_ON                 0x39
#define LCD6100_COLOR_INTERFACE         0x3A
  // pixelformat                0xXX
  #define LCD6100_COLOR_8_BIT           0x02
  #define LCD6100_COLOR_12_BIT          0x03
  #define LCD6100_COLOR_16_BIT          0x05
#define LCD6100_SET_VOP                 0xB0  // irrelevant for Nokia6100
  // 2x                         0xXX
#define LCD6100_BOTTOM_ROW_SWAP_OFF     0xB4
#define LCD6100_BOTTOM_ROW_SWAP_ON      0xB5
#define LCD6100_TOP_ROW_SWAP_OFF        0xB6
#define LCD6100_TOP_ROW_SWAP_ON         0xB7
#define LCD6100_SUPER_FRAME_INV_OFF     0xB8
#define LCD6100_SUPER_FRAME_INV_ON      0xB9
#define LCD6100_DATA_ORDER_MSB          0xBA
#define LCD6100_DATA_ORDER_LSB          0xBB
#define LCD6100_TEMP_READ               0xC8  // read
  // temp                       0xXX


void   InitLCD6100(void);
UINT16 LCD6100_Print(UINT8 far *addr);
void   LCD6100_NewLine(void);
void   LCD6100_FillRect(LCD6100_Coord_t x1, LCD6100_Coord_t y1, LCD6100_Coord_t x2, LCD6100_Coord_t y2, const LCD6100_Color_t color);
void   LCD6100_SetPixel(LCD6100_Coord_t x, LCD6100_Coord_t y, const LCD6100_Color_t color);

#define FALSE 0
#define TRUE 1

#endif